#!/usr/bin/env python3

"""
This Python script is designed to query multiple online repositories for the
synonyms associated with the given word.
"""
__author__ = 'John Bumgarner'
__date__ = 'October 15, 2020'
__status__ = 'Production'
__license__ = 'MIT'
__copyright__ = "Copyright (C) 2020 John Bumgarner"

##################################################################################
# Date Completed: October 15, 2020
# Author: John Bumgarner
#
# Date Last Revised: July 3, 2021
# Revised by: John Bumgarner
##################################################################################

##################################################################################
# “AS-IS” Clause
#
# Except as represented in this agreement, all work produced by Developer is
# provided “AS IS”. Other than as provided in this agreement, Developer makes no
# other warranties, express or implied, and hereby disclaims all implied warranties,
# including any warranty of merchantability and warranty of fitness for a particular
# purpose.
##################################################################################

##################################################################################
# Python imports required for basic operations
##################################################################################
import bs4
import logging
import requests
import traceback
import re as regex
from bs4 import BeautifulSoup
from wordhoard.utilities import basic_soup, caching, cleansing, word_verification

logger = logging.getLogger(__name__)


class Synonyms(object):
    """
    This class is used to query multiple online repositories for the synonyms associated
    with a specific word.

    Usage: synonym = Synonyms(word)
           results = synonym.find_synonyms()
    """

    def __init__(self, word):
        """
        :param word: string containing the variable to search for
        """
        self._word = word

    def _validate_word(self):
        """
        This function is designed to validate that the syntax for
        a string variable is in an acceptable format.

        :return: True or False
        :rtype: bool
        """
        valid_word = word_verification.validate_word_syntax(self._word)
        if valid_word:
            return valid_word
        else:
            logger.error(f'The word {self._word} was not in a valid format.')
            logger.error(f'Please verify that the word {self._word} is spelled correctly.')

    def _check_cache(self):
        check_cache = caching.cache_synonyms(self._word)
        return check_cache

    def _update_cache(self, synonyms):
        caching.insert_word_cache_synonyms(self._word, synonyms)
        return

    def find_synonyms(self):
        """
        This function queries multiple online repositories to discover synonyms
        associated with the specific word provided to the Class Synonyms.
        The synonyms are deduplicated and sorted alphabetically.

        :returns:
            synonyms: list of synonyms

        :rtype: list
        """
        valid_word = self._validate_word()
        if valid_word:
            check_cache = self._check_cache()
            if check_cache is False:
                synonyms_01 = self._query_collins_dictionary()
                synonyms_02 = self._query_synonym_com()
                synonyms_03 = self._query_thesaurus_plus()
                synonyms_04 = self._query_wordnet()
                synonyms = ([x for x in [synonyms_01, synonyms_02, synonyms_03, synonyms_04] if x is not None])
                synonyms_results = cleansing.flatten_multidimensional_list(synonyms)
                return sorted(set(synonyms_results))
            else:
                synonyms = cleansing.flatten_multidimensional_list([val for val in check_cache.values()])
                return synonyms

    def _query_collins_dictionary(self):
        """
        This function queries collinsdictionary.com for synonyms associated
        with the specific word provided to the Class Synonyms.

         :returns:
            synonyms: list of synonyms

        :rtype: list

        :raises
            AttributeError: Raised when an attribute reference or assignment fails.

            KeyError: Raised when a mapping (dictionary) key is not found in the set of existing keys.

            TypeError: Raised when an operation or function is applied to an object of inappropriate type.

            bs4.FeatureNotFound: raised by the BeautifulSoup constructor if no parser with the requested features
            is found
        """
        try:
            synonyms = []
            results_synonym = basic_soup.get_single_page_html(
                f'https://www.collinsdictionary.com/dictionary/english-thesaurus/{self._word}')
            soup = BeautifulSoup(results_synonym, 'lxml')
            word_found = soup.find('h1', text=f'Sorry, no results for “{self._word}” in the English Thesaurus.')
            if word_found:
                logger.error(f'Collins Dictionary had no reference for the word {self._word}')
                logger.error(f'Please verify that the word {self._word} is spelled correctly.')
            else:
                query_results = basic_soup.query_html(results_synonym, 'div', 'class', 'blockSyn')
                content_descendants = query_results.descendants
                for item in content_descendants:
                    if item.name == 'div' and item.get('class', 'form type-syn orth'):
                        children = item.findChild('span', {'class': 'orth'})
                        if children is not None:
                            synonyms.append(children.text)

                self._update_cache(synonyms)
                return sorted(synonyms)
        except bs4.FeatureNotFound as error:
            logger.error('A BeautifulSoup error occurred in the following code segment:')
            logger.error(''.join(traceback.format_tb(error.__traceback__)))
        except AttributeError as error:
            logger.error('An AttributeError occurred in the following code segment:')
            logger.error(''.join(traceback.format_tb(error.__traceback__)))
        except KeyError as error:
            logger.error('A KeyError occurred in the following code segment:')
            logger.error(''.join(traceback.format_tb(error.__traceback__)))
        except TypeError as error:
            logger.error('A TypeError occurred in the following code segment:')
            logger.error(''.join(traceback.format_tb(error.__traceback__)))

    def _query_synonym_com(self):
        """
        This function queries synonym.com for synonyms associated
        with the specific word provided to the Class Synonyms.

         :returns:
            synonyms: list of synonyms

        :rtype: list

        :raises
            AttributeError: Raised when an attribute reference or assignment fails.

            IndexError: Raised when a sequence subscript is out of range

            KeyError: Raised when a mapping (dictionary) key is not found in the set of existing keys.

            TypeError: Raised when an operation or function is applied to an object of inappropriate type.

            bs4.FeatureNotFound: raised by the BeautifulSoup constructor if no parser with the requested features
            is found
        """
        try:
            results_synonym = basic_soup.get_single_page_html(
                f'https://www.synonym.com/synonyms/{self._word}')
            soup = BeautifulSoup(results_synonym, "lxml")
            description_tag = soup.find("meta", property="og:description")
            if 'find any words based on your search' in description_tag['content']:
                logger.error(f'synonym.com had no reference for the word {self._word}')
                logger.error(f'Please verify that the word {self._word} is spelled correctly.')
            else:
                find_synonyms = regex.split(r'\|', description_tag['content'])
                synonyms_list = find_synonyms[2].lstrip().replace('synonyms:', '').split(',')
                synonyms = [cleansing.normalize_space(i) for i in synonyms_list]
                self._update_cache(synonyms)
                return sorted(synonyms)
        except bs4.FeatureNotFound as error:
            logger.error('An error occurred in the following code segment:')
            logger.error(''.join(traceback.format_tb(error.__traceback__)))
        except AttributeError as error:
            logger.error('An AttributeError occurred in the following code segment:')
            logger.error(''.join(traceback.format_tb(error.__traceback__)))
        except IndexError as error:
            logger.error('An IndexError occurred in the following code segment:')
            logger.error(''.join(traceback.format_tb(error.__traceback__)))
        except KeyError as error:
            logger.error('A KeyError occurred in the following code segment:')
            logger.error(''.join(traceback.format_tb(error.__traceback__)))
        except TypeError as error:
            logger.error('A TypeError occurred in the following code segment:')
            logger.error(''.join(traceback.format_tb(error.__traceback__)))

    def _query_thesaurus_plus(self):
        """
        This function queries thesaurus.plus for synonyms associated
        with the specific word provided to the Class Synonyms.

        :returns:
            synonyms: list of synonyms

        :rtype: list

        :raises

            IndexError: Raised when a sequence subscript is out of range.

            requests.ConnectionError: Raised when a connection error has occurred.

            requests.HTTPError: Raised when an HTTP error has occurred.

            requests.RequestException: Raised when an unknown error has occurred.

            requests.Timeout: Raised when the request timed out.

        """
        try:
            synonyms_list = []
            results_synonym = basic_soup.get_single_page_html(
                f'https://thesaurus.plus/synonyms/{self._word}/category/noun')
            soup = BeautifulSoup(results_synonym, "lxml")
            no_word = soup.find('title', text='404. Page not found')
            if no_word:
                logger.error(f'thesaurus.plus has no reference for the word {self._word}')
                logger.error(f'Please verify that the word {self._word} is spelled correctly.')
            else:
                synonyms = []
                parent_node = soup.find('ul', {'class': 'list paper'}).findAll('li')[1:]
                for children in parent_node:
                    for child in children.findAll('div', {'class': 'action_pronounce'}):
                        split_dictionary = str(child.attrs).split(',')
                        synonyms_list.append(split_dictionary[1].replace("'data-term':", "").replace("'", ""))
                        synonyms = sorted([cleansing.normalize_space(i) for i in synonyms_list])
                self._update_cache(synonyms)
                return synonyms
        except IndexError as error:
            logger.error('A IndexError occurred in the following code segment:')
            logger.error(''.join(traceback.format_tb(error.__traceback__)))
        except requests.HTTPError as error:
            logger.error('A HTTP error has occurred.')
            logger.error(''.join(traceback.format_tb(error.__traceback__)))
        except requests.ConnectionError as error:
            if requests.codes:
                'Failed to establish a new connection'
                logger.error(''.join(traceback.format_tb(error.__traceback__)))
        except requests.Timeout as error:
            logger.error('A connection timeout has occurred.')
            logger.error(''.join(traceback.format_tb(error.__traceback__)))
        except requests.RequestException as error:
            logger.error('An ambiguous exception occurred.')
            logger.error(''.join(traceback.format_tb(error.__traceback__)))

    def _query_wordnet(self):
        """
        This function queries wordnet for synonyms associated
        with the specific word provided to the Class Synonyms.

        :returns:
            synonyms: list of synonyms

        :rtype: list

        :raises

            IndexError: Raised when a sequence subscript is out of range.

            requests.ConnectionError: Raised when a connection error has occurred.

            requests.HTTPError: Raised when an HTTP error has occurred.

            requests.RequestException: Raised when an unknown error has occurred.

            requests.Timeout: Raised when the request timed out.

        """
        try:
            synonyms = []
            results = requests.get(f'http://wordnetweb.princeton.edu/perl/webwn?s={self._word}',
                                   headers=basic_soup.http_headers,
                                   allow_redirects=True, verify=True, timeout=30)
            soup = BeautifulSoup(results.text, "lxml")
            if soup.findAll('h3', text='Noun'):
                parent_node = soup.findAll("ul")[0].findAll('li')
                for children in parent_node:
                    for child in children.find_all(href=True):
                        if 'S:' not in child.contents[0]:
                            synonyms.append(child.contents[0])
                synonyms = sorted([x.lower() for x in synonyms])
                self._update_cache(synonyms)
                return synonyms
            else:
                logger.error(f'Wordnet had no reference for the word {self._word}')
                logger.error(f'Please verify that the word {self._word} is spelled correctly.')
        except IndexError as error:
            logger.error('A IndexError occurred in the following code segment:')
            logger.error(''.join(traceback.format_tb(error.__traceback__)))
        except requests.HTTPError as error:
            logger.error('A HTTP error has occurred.')
            logger.error(''.join(traceback.format_tb(error.__traceback__)))
        except requests.ConnectionError as error:
            if requests.codes:
                'Failed to establish a new connection'
                logger.error(''.join(traceback.format_tb(error.__traceback__)))
        except requests.Timeout as error:
            logger.error('A connection timeout has occurred.')
            logger.error(''.join(traceback.format_tb(error.__traceback__)))
        except requests.RequestException as error:
            logger.error('An ambiguous exception occurred.')
            logger.error(''.join(traceback.format_tb(error.__traceback__)))



#!/usr/bin/env python3

"""
This Python script is designed to query internal repositories for the
homophones associated with the given word.
"""
__author__ = 'John Bumgarner'
__date__ = 'June 11, 2021'
__status__ = 'Production'
__license__ = 'MIT'
__copyright__ = "Copyright (C) 2021 John Bumgarner"

##################################################################################
# Date Completed: June 11, 2021
# Author: John Bumgarner
#
# Date Last Revised: July 4, 2021
# Revised by: John Bumgarner
##################################################################################

##################################################################################
# “AS-IS” Clause
#
# Except as represented in this agreement, all work produced by Developer is
# provided “AS IS”. Other than as provided in this agreement, Developer makes no
# other warranties, express or implied, and hereby disclaims all implied warranties,
# including any warranty of merchantability and warranty of fitness for a particular
# purpose.
##################################################################################

##################################################################################
# Python imports required for basic operations
##################################################################################
import os
import sys
import pickle
import logging
import traceback
from wordhoard.utilities import word_verification

logger = logging.getLogger(__name__)

PARENT_DIRECTORY = os.path.dirname(os.path.abspath(__file__))

# Opening the pickle file that contains a nested list of common
# English language homophones.
try:
    _file_known_homophones = os.path.join(PARENT_DIRECTORY, 'files/common_english_homophones.pkl')
    with open(_file_known_homophones, 'rb') as _eng_homophones:
        _known_homophones_list = pickle.load(_eng_homophones)
        _eng_homophones.close()
except FileNotFoundError as error:
    logger.error('The common_english_homophones.pkl file was not found. Aborting operation.')
    logger.error(''.join(traceback.format_tb(error.__traceback__)))
    sys.exit(1)
except OSError as error:
    logger.error(f"An OS error occurred when trying to open the file common_english_homophones.pkl")
    logger.error(''.join(traceback.format_tb(error.__traceback__)))
    sys.exit(1)


# Opening the pickle file that contains a nested list of English
# language words that have no known homophones.
try:
    _file_no_known_homophones = os.path.join(PARENT_DIRECTORY, 'files/no_homophones_english.pkl')
    with open(_file_no_known_homophones, 'rb') as _no_eng_homophones:
        _no_homophones_list = pickle.load(_no_eng_homophones)
        _no_eng_homophones.close()
except FileNotFoundError as error:
    logger.error('The no_homophones_english.pkl file was not found. Aborting operation.')
    logger.error(''.join(traceback.format_tb(error.__traceback__)))
    sys.exit(1)
except OSError as error:
    logger.error(f"An OS error occurred when trying to open the file no_homophones_english.pkl")
    logger.error(''.join(traceback.format_tb(error.__traceback__)))
    sys.exit(1)


class Homophones(object):
    """
    This class is used to query internal files containing
    English language homophones associated with a specific word.

    Usage: homophone = Homophones(word)
           results = homophone.find_homophones()
    """

    def __init__(self, word):
        """
        :param word: string containing the variable to search for
        """
        self._word = word

    def _validate_word(self):
        """
        This function is designed to validate that the syntax for
        a string variable is in an acceptable format.

        :return: True or False
        :rtype: boolean
        """
        valid_word = word_verification.validate_word_syntax(self._word)
        if valid_word:
            return valid_word
        else:
            logger.error(f'The word {self._word} was not in a valid format.')
            logger.error(f'Please verify that the word {self._word} is spelled correctly.')

    def _common_english_homophones(self):
        """
        This function iterates through a list of known
        English language homophones.

        :return:
            homophones: list of homophones

        :rtype: list
        """
        global _known_homophones_list
        rtn_list = []
        for homophones in _known_homophones_list:
            match = bool([word for word in homophones if word == self._word])
            if match:
                for word in homophones:
                    if word != self._word:
                        rtn_list.append(f'{self._word} is a homophone of {word}')
        if len(rtn_list) > 0:
            return list(set(rtn_list))

    def _english_words_without_homophones(self):
        """
        This function iterates through a list of English
        language words with no known homophones.

        :return: no homophones for word
        :rtype: str
        """
        global _no_homophones_list
        match = bool(self._word in _no_homophones_list)
        if match:
            return f'no homophones for {self._word}'

    def find_homophones(self):
        """
        This function queries multiple lists to find an
        English language homophones associated with the
        specific word provided to the Class Homophones.

        :return:
            homophones: list of homophones
            no_homophones: string

        :rtype: list
        :rtype: str
        """
        valid_word = self._validate_word()
        if valid_word:
            known_english_homophones = self._common_english_homophones()
            if known_english_homophones:
                return known_english_homophones
            elif not known_english_homophones:
                return self._english_words_without_homophones()



#!/usr/bin/env python3

"""
This Python script is designed to query a online repository for the
hypernyms associated with a specific word.
"""
__author__ = 'John Bumgarner'
__date__ = 'June 12, 2021'
__status__ = 'Production'
__license__ = 'MIT'
__copyright__ = "Copyright (C) 2021 John Bumgarner"

##################################################################################
# Date Initially Completed: June 12, 2021
# Author: John Bumgarner
#
# Date Last Revised: July 4, 2021
# Revised by: John Bumgarner
##################################################################################


##################################################################################
# “AS-IS” Clause
#
# Except as represented in this agreement, all work produced by Developer is
# provided “AS IS”. Other than as provided in this agreement, Developer makes no
# other warranties, express or implied, and hereby disclaims all implied warranties,
# including any warranty of merchantability and warranty of fitness for a particular
# purpose.
##################################################################################

##################################################################################
# Python imports required for basic operations
##################################################################################
import bs4
import logging
import requests
import traceback
from bs4 import BeautifulSoup
from wordhoard.utilities import basic_soup, caching, cleansing, word_verification

logger = logging.getLogger(__name__)


def _get_number_of_pages(soup):
    """
    This function determines the number of pages that
    contain hypernyms and hyperonyms for a specific word.

    :param soup: BeautifulSoup lxml
    :return: number of pages
    :rtype: int

    :raises

        AttributeError: Raised when an attribute reference or assignment fails.

        KeyError: Raised when a mapping (dictionary) key is not found in the set of existing keys.

        TypeError: Raised when an operation or function is applied to an object of inappropriate type.
    """
    try:
        number_of_pages = 0
        pages = soup.find('div', {'id': 'pages'})
        if pages is not None:
            list_of_pages = [num for page in pages for num in page if num.isdigit()]
            if len(list_of_pages) != 0:
                number_of_pages = int(list_of_pages[-1]) + 1
        return number_of_pages

    except AttributeError as e:
        logger.error('An AttributeError occurred in the following code segment:')
        logger.error(''.join(traceback.format_tb(e.__traceback__)))
    except KeyError as e:
        logger.error('A KeyError occurred in the following code segment:')
        logger.error(''.join(traceback.format_tb(e.__traceback__)))
    except TypeError as e:
        logger.error('A TypeError occurred in the following code segment:')
        logger.error(''.join(traceback.format_tb(e.__traceback__)))


def _get_hypernyms(soup):
    """
    This function queries a HTML table for hypernyms.

    :param soup: BeautifulSoup lxml
    :return: set of hypernyms and hyperonyms
    :rtype: set

    :raises

        AttributeError: Raised when an attribute reference or assignment fails.

        KeyError: Raised when a mapping (dictionary) key is not found in the set of existing keys.

        TypeError: Raised when an operation or function is applied to an object of inappropriate type.
    """
    try:
        sub_set = set()
        table = soup.find('table')
        if table:
            rows = table.find_all('tr', {'class': 'theentry'})
            if rows is not None:
                for row in rows:
                    cols = row.find('td', {'class': 'abbdef'}).find('a')
                    if cols is not None:
                        if cols.text != '»':
                            sub_set.add(str(cols.text).lower())
                        else:
                            sub_set.add('no hypernyms found')
            return sub_set

    except AttributeError as e:
        logger.error('An AttributeError occurred in the following code segment:')
        logger.error(''.join(traceback.format_tb(e.__traceback__)))
    except KeyError as e:
        logger.error('A KeyError occurred in the following code segment:')
        logger.error(''.join(traceback.format_tb(e.__traceback__)))
    except TypeError as e:
        logger.error('A TypeError occurred in the following code segment:')
        logger.error(''.join(traceback.format_tb(e.__traceback__)))


class Hypernyms(object):
    """
    This class is used to query online repositories for the hypernyms associated
    with a specific word.

    Usage:
      hypernym = Hypernyms(word)

      results = hypernym.find_hypernyms

    """

    def __init__(self, word):
        """
        :param word: string variable used to find hypernyms for
        """
        self._word = word

    def _validate_word(self):
        """
        This function is designed to validate that the syntax for
        a string variable is in an acceptable format.

        :return: True or False
        :rtype bool

        """
        valid_word = word_verification.validate_word_syntax(self._word)
        if valid_word:
            return valid_word
        else:
            logger.error(f'The word {self._word} was not in a valid format.')
            logger.error(f'Please verify that the word {self._word} is spelled correctly.')

    def _check_cache(self):
        check_cache = caching.cache_hypernyms(self._word)
        return check_cache

    def _update_cache(self, hypernym):
        caching.insert_word_cache_hypernyms(self._word, hypernym)
        return

    def find_hypernyms(self):
        """
        This function queries classicthesaurus_com for hypernyms associated
        with the specific word provided to the Class Hypernyms.

        :returns:
            hypernym: list of hypernyms

        :rtype: list

        :raises

            AttributeError: Raised when an attribute reference or assignment fails.

            KeyError: Raised when a mapping (dictionary) key is not found in the set of existing keys.

            TypeError: Raised when an operation or function is applied to an object of inappropriate type.

            bs4.FeatureNotFound: raised by the BeautifulSoup constructor if no parser with the requested features
            is found
        """
        valid_word = self._validate_word()
        if valid_word:
            check_cache = self._check_cache()
            if check_cache is not False:
                hypernym = cleansing.flatten_multidimensional_list([val for val in check_cache.values()])
                return hypernym
            elif check_cache is False:
                try:
                    results_hypernyms = basic_soup.get_single_page_html(
                        f'https://www.classicthesaurus.com/{self._word}/broader')
                    soup = BeautifulSoup(results_hypernyms, "lxml")
                    hypernym = _get_hypernyms(soup)

                    number_of_pages = _get_number_of_pages(soup)
                    if number_of_pages >= 2:
                        for page in range(2, number_of_pages):
                            sub_html = requests.get(f'https://www.classicthesaurus.com/{self._word}/broader/{page}',
                                                    basic_soup.http_headers)
                            sub_soup = BeautifulSoup(sub_html.text, 'lxml')
                            additional_hypernym = _get_hypernyms(sub_soup)
                            if additional_hypernym:
                                hypernym.union(additional_hypernym)

                    self._update_cache(hypernym)
                    return sorted(hypernym)

                except bs4.FeatureNotFound as e:
                    logger.error('An error occurred in the following code segment:')
                    logger.error(''.join(traceback.format_tb(e.__traceback__)))
                except AttributeError as e:
                    logger.error('An AttributeError occurred in the following code segment:')
                    logger.error(''.join(traceback.format_tb(e.__traceback__)))
                except KeyError as e:
                    logger.error('A KeyError occurred in the following code segment:')
                    logger.error(''.join(traceback.format_tb(e.__traceback__)))
                except TypeError as e:
                    logger.error('A TypeError occurred in the following code segment:')
                    logger.error(''.join(traceback.format_tb(e.__traceback__)))



#!/usr/bin/env python3

"""
This Python script is designed to query multiple online repositories for the
definition associated with a given word.
"""
__author__ = 'John Bumgarner'
__date__ = 'October 15, 2020'
__status__ = 'Production'
__license__ = 'MIT'
__copyright__ = "Copyright (C) 2021 John Bumgarner"

##################################################################################
# Date Completed: October 15, 2020
# Author: John Bumgarner
#
# Date Last Revised: July 3, 2021
# Revised by: John Bumgarner
##################################################################################

##################################################################################
# “AS-IS” Clause
#
# Except as represented in this agreement, all work produced by Developer is
# provided “AS IS”. Other than as provided in this agreement, Developer makes no
# other warranties, express or implied, and hereby disclaims all implied warranties,
# including any warranty of merchantability and warranty of fitness for a particular
# purpose.
##################################################################################

##################################################################################
# Python imports required for basic operations
##################################################################################
import bs4
import logging
import requests
import traceback
import re as regex
from bs4 import BeautifulSoup
from wordhoard.utilities import basic_soup, caching, cleansing, word_verification

logger = logging.getLogger(__name__)


class Definitions:
    """
    This class is used to query multiple online repositories for the
    definition associated with a specific word.

    Usage:
      definition = Definitions(word)
      results = definition.find_definitions()
    """

    def __init__(self, search_string):
        """
        :param search_string: string containing the variable to search for
        """
        self._word = search_string

    def _validate_word(self):
        """
        This function is designed to validate that the syntax for
        a string variable is in an acceptable format.

        :return: True or False
        :rtype: bool
        """
        valid_word = word_verification.validate_word_syntax(self._word)
        if valid_word:
            return valid_word
        else:
            logger.error(f'The word {self._word} was not in a valid format.')
            logger.error(f'Please verify that the word {self._word} is spelled correctly.')

    def _update_cache(self, definition):
        caching.insert_word_cache_definition(self._word, definition)
        return

    def find_definitions(self):
        """
        This function queries multiple online repositories to discover
        definitions related with the specific word provided to the
        Class Definitions.

        :return: list of definitions
        :rtype: list
        """
        valid_word = self._validate_word()
        if valid_word:
            check_cache = caching.cache_antonyms(self._word)
            if check_cache is False:
                definition_01 = self._query_synonym_com()
                definition_02 = self._query_collins_dictionary()
                definition_03 = self._query_thesaurus_com()
                definitions = ([x for x in [definition_01, definition_02, definition_03] if x is not None])
                return sorted(set(definitions))
            else:
                definitions = cleansing.flatten_multidimensional_list([val for val in check_cache.values()])
                return definitions

    def _query_synonym_com(self):
        """
        This function queries synonym_com for a definition associated
        with the specific word provided to the Class Definitions.

         :returns:
            definition: definition for a word

        :rtype: string

        :raises
            AttributeError: Raised when an attribute reference or assignment fails.

            KeyError: Raised when a mapping (dictionary) key is not found in the set of existing keys.

            TypeError: Raised when an operation or function is applied to an object of inappropriate type.

            bs4.FeatureNotFound: raised by the BeautifulSoup constructor if no parser with the requested features
            is found
        """
        try:
            results_definition = basic_soup.get_single_page_html(
                f'https://www.synonym.com/synonyms/{self._word}')
            soup = BeautifulSoup(results_definition, "lxml")
            description_tag = soup.find("meta", property="og:description")
            if 'find any words based on your search' not in description_tag['content']:
                find_definition = regex.split(r'\|', description_tag['content'])
                definition_list = find_definition[1].lstrip().replace('definition:', '').split(',')
                definition = [cleansing.normalize_space(i) for i in definition_list]
                definition_list_to_string = ' '.join([str(elem) for elem in definition])
                self._update_cache(definition_list_to_string)
                return definition_list_to_string
            else:
                logger.error(f'synonym.com had no reference for the word {self._word}')
        except bs4.FeatureNotFound as error:
            logger.error('An error occurred in the following code segment:')
            logger.error(''.join(traceback.format_tb(error.__traceback__)))
        except AttributeError as error:
            logger.error('An AttributeError occurred in the following code segment:')
            logger.error(''.join(traceback.format_tb(error.__traceback__)))
        except KeyError as error:
            logger.error('A KeyError occurred in the following code segment:')
            logger.error(''.join(traceback.format_tb(error.__traceback__)))
        except TypeError as error:
            logger.error('A TypeError occurred in the following code segment:')
            logger.error(''.join(traceback.format_tb(error.__traceback__)))

    def _query_collins_dictionary(self):
        """
        This function queries collinsdictionary.com for a definition associated
        with the specific word provided to the Class Definitions.

         :returns:
            definition: definition for a word

        :rtype: str

        :raises
            AttributeError: Raised when an attribute reference or assignment fails.

            KeyError: Raised when a mapping (dictionary) key is not found in the set of existing keys.

            TypeError: Raised when an operation or function is applied to an object of inappropriate type.

            bs4.FeatureNotFound: raised by the BeautifulSoup constructor if no parser with the requested features
            is found
        """
        try:
            results_definition = basic_soup.get_single_page_html(
                f'https://www.collinsdictionary.com/dictionary/english-thesaurus/{self._word}')
            query_results = basic_soup.query_html(results_definition, 'div', 'class',
                                                  'form type-def titleTypeSubContainer')
            if query_results is not None:
                definition = query_results.findNext('div', {'class': 'def'})
                self._update_cache(definition.text)
                return definition.text
            else:
                logger.error(f'Collins Dictionary had no reference for the word {self._word}')
        except bs4.FeatureNotFound as error:
            logger.error('An error occurred in the following code segment:')
            logger.error(''.join(traceback.format_tb(error.__traceback__)))
        except AttributeError as error:
            logger.error('An AttributeError occurred in the following code segment:')
            logger.error(''.join(traceback.format_tb(error.__traceback__)))
        except KeyError as error:
            logger.error('A KeyError occurred in the following code segment:')
            logger.error(''.join(traceback.format_tb(error.__traceback__)))
        except TypeError as error:
            logger.error('A TypeError occurred in the following code segment:')
            logger.error(''.join(traceback.format_tb(error.__traceback__)))

    def _query_thesaurus_com(self):
        """
        This function queries thesaurus.com for a definition associated
        with the specific word provided to the Class Definitions.

        :returns:
            definition: definition for a word

        :rtype: str

        :raises

            IndexError: Raised when a sequence subscript is out of range.

            requests.ConnectionError: Raised when a connection error has occurred.

            requests.HTTPError: Raised when an HTTP error has occurred.

            requests.RequestException: Raised when an unknown error has occurred.

            requests.Timeout: Raised when the request timed out.
        """
        try:
            req = requests.get(f'https://tuna.thesaurus.com/pageData/{self._word}',
                               headers=basic_soup.http_headers,
                               allow_redirects=True, verify=True, timeout=30)
            if req.json()['data'] is not None:
                definition = req.json()['data']['definitionData']['definitions'][0]['definition']
                self._update_cache(definition)
                return definition
            else:
                logger.error(f'thesaurus.com had no reference for the word {self._word}')
        except requests.HTTPError as error:
            logger.error('A HTTP error has occurred.')
            logger.error(''.join(traceback.format_tb(error.__traceback__)))
        except requests.ConnectionError as error:
            if requests.codes:
                'Failed to establish a new connection'
                logger.error(''.join(traceback.format_tb(error.__traceback__)))
        except requests.Timeout as error:
            logger.error('A connection timeout has occurred.')
            logger.error(''.join(traceback.format_tb(error.__traceback__)))
        except requests.RequestException as error:
            logger.error('An ambiguous exception occurred.')
            logger.error(''.join(traceback.format_tb(error.__traceback__)))



__author__ = 'John Bumgarner'
__date__ = 'July 5, 2021'
__status__ = 'Production'
__license__ = 'MIT'
__copyright__ = "Copyright (C) 2021 John Bumgarner"

from .antonyms import Antonyms
from .synonyms import Synonyms
from .hyponyms import Hyponyms
from .hypernyms import Hypernyms
from .homophones import Homophones
from .dictionary import Definitions

import logging
from .utilities import wordhoard_logger

logger = logging.getLogger(__name__)
wordhoard_logger.enable_logging(logger)



import setuptools

with open("PYPI_description.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="wordhoard",
    version="1.4.6",
    author="John Bumgarner",
    author_email="wordhoardproject@gmail.com",
    description="A comprehensive lexical discovery application that is useful for finding semantic relationships between "
                "words including a word's synonyms, synonyms, hypernyms, hyponyms and homophones.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/johnbumgarner/wordhoard",
    packages=setuptools.find_packages(),
    include_package_data=True,
    package_data={'files': ['files/common_english_homophones.pkl',
                            'files/no_homophones_english.pkl']},
    license='LICENSE.txt',
    classifiers=["Development Status :: 5 - Production/Stable",
                 "Intended Audience :: Developers",
                 "License :: OSI Approved :: MIT License",
                 "Natural Language :: English",
                 "Operating System :: OS Independent",
                 "Programming Language :: Python :: 3.6",
                 "Programming Language :: Python :: 3.7",
                 "Programming Language :: Python :: 3.8",
                 "Programming Language :: Python :: 3.9",
                 "Topic :: Software Development :: Libraries",
                 "Topic :: Text Processing :: Linguistic",
                 "Topic :: Utilities"],
    keywords=['antonyms', 'synonyms', 'hypernyms', 'hyponyms', 'homophones', 'definitions', 'lexicon',
              'semantic relationships', 'natural language processing'],
    python_requires='>=3.6',
    install_requires=['bs4',
                      'deep-translator',
                      'lxml',
                      'requests',
                      'urllib3']
)




#!/usr/bin/env python3

"""
This Python script is designed to query a online repository for the
hyponyms associated with a specific word.
"""
__author__ = 'John Bumgarner'
__date__ = 'June 12, 2021'
__status__ = 'Production'
__license__ = 'MIT'
__copyright__ = "Copyright (C) 2021 John Bumgarner"

##################################################################################
# Date Initially Completed: June 12, 2021
# Author: John Bumgarner
#
# Date Last Revised: July 4, 2021
# Revised by: John Bumgarner
###################################################################################

##################################################################################
# “AS-IS” Clause
#
# Except as represented in this agreement, all work produced by Developer is
# provided “AS IS”. Other than as provided in this agreement, Developer makes no
# other warranties, express or implied, and hereby disclaims all implied warranties,
# including any warranty of merchantability and warranty of fitness for a particular
# purpose.
##################################################################################

##################################################################################
# Python imports required for basic operations
##################################################################################
import bs4
import logging
import requests
import traceback
from bs4 import BeautifulSoup
from wordhoard.utilities import basic_soup, caching, cleansing, wordhoard_logger, word_verification

logger = logging.getLogger(__name__)
wordhoard_logger.enable_logging(logger)


def _get_number_of_pages(soup):
    """
    This function determines the number of pages
    that contain hyponyms for a specific word.

    :param soup: BeautifulSoup lxml

    :return: number of pages

    :rtype: int

    :raises
        AttributeError: Raised when an attribute reference or assignment fails.

        KeyError: Raised when a mapping (dictionary) key is not found in the set of existing keys.

        TypeError: Raised when an operation or function is applied to an object of inappropriate type.

    """
    try:
        number_of_pages = 0
        pages = soup.find('div', {'id': 'pages'})
        if pages is not None:
            list_of_pages = [num for page in pages for num in page if num.isdigit()]
            if len(list_of_pages) != 0:
                number_of_pages = int(list_of_pages[-1]) + 1
        return number_of_pages

    except AttributeError as error:
        logger.error('An AttributeError occurred in the following code segment:')
        logger.error(''.join(traceback.format_tb(error.__traceback__)))
    except KeyError as error:
        logger.error('A KeyError occurred in the following code segment:')
        logger.error(''.join(traceback.format_tb(error.__traceback__)))
    except TypeError as error:
        logger.error('A TypeError occurred in the following code segment:')
        logger.error(''.join(traceback.format_tb(error.__traceback__)))


def _get_hyponyms(soup):
    """
    This function queries a HTML table for hyponyms.

    :param soup: BeautifulSoup lxml

    :return:
        hyponyms: set of hyponyms

    :rtype: set

    :raises
        AttributeError: Raised when an attribute reference or assignment fails.

        KeyError: Raised when a mapping (dictionary) key is not found in the set of existing keys.

        TypeError: Raised when an operation or function is applied to an object of inappropriate type.
    """
    try:
        sub_set = set()
        table = soup.find('table')
        if table:
            rows = table.find_all('tr', {'class': 'theentry'})
            if rows is not None:
                for row in rows:
                    cols = row.find('td', {'class': 'abbdef'}).find('a')
                    if cols is not None:
                        if cols.text != '»':
                            sub_set.add(str(cols.text).lower())
                        else:
                            sub_set.add('no hyponyms found')
        return sub_set

    except AttributeError as error:
        logger.error('An AttributeError occurred in the following code segment:')
        logger.error(''.join(traceback.format_tb(error.__traceback__)))
    except KeyError as error:
        logger.error('A KeyError occurred in the following code segment:')
        logger.error(''.join(traceback.format_tb(error.__traceback__)))
    except TypeError as error:
        logger.error('A TypeError occurred in the following code segment:')
        logger.error(''.join(traceback.format_tb(error.__traceback__)))


class Hyponyms(object):
    """
    This class is used to query online repositories for the hyponyms associated
    with a specific word.

    Usage:
      hyponym = Hyponyms(word)
      results = hyponyms.find_hyponyms()
    """

    def __init__(self, word):
        """
        :param word: string variable used to find hyponyms for
        """
        self._word = word

    def _validate_word(self):
        """
        This function is designed to validate that the syntax for
        a string variable is in an acceptable format.

        :return: True or False
        :rtype: bool
        """
        valid_word = word_verification.validate_word_syntax(self._word)
        if valid_word:
            return valid_word
        else:
            logger.error(f'The word {self._word} was not in a valid format.')
            logger.error(f'Please verify that the word {self._word} is spelled correctly.')

    def _check_cache(self):
        check_cache = caching.cache_hyponyms(self._word)
        return check_cache

    def _update_cache(self, hyponyms):
        caching.insert_word_cache_hyponyms(self._word, hyponyms)
        return

    def find_hyponyms(self):
        """
        This function queries classicthesaurus_com for hyponyms associated
        with the specific word provided to the Class Hyponyms.

         :returns:
             hyponyms: list of hyponyms

        :rtype: list

        :raises
            AttributeError: Raised when an attribute reference or assignment fails.

            IndexError: Raised when a sequence subscript is out of range

            KeyError: Raised when a mapping (dictionary) key is not found in the set of existing keys.

            TypeError: Raised when an operation or function is applied to an object of inappropriate type.

            bs4.FeatureNotFound: raised by the BeautifulSoup constructor if no parser with the requested features
            is found
        """
        valid_word = self._validate_word()
        if valid_word:
            check_cache = self._check_cache()
            if check_cache is False:
                try:
                    results_hyponyms = basic_soup.get_single_page_html(
                        f'https://www.classicthesaurus.com/{self._word}/narrower')
                    soup = BeautifulSoup(results_hyponyms, "lxml")
                    hyponym = _get_hyponyms(soup)

                    number_of_pages = _get_number_of_pages(soup)
                    if number_of_pages >= 2:
                        for page in range(2, number_of_pages):
                            sub_html = requests.get(f'https://www.classicthesaurus.com/{self._word}/narrower/{page}',
                                                    headers=basic_soup.http_headers)
                            sub_soup = BeautifulSoup(sub_html.text, 'lxml')
                            additional_hyponym = _get_hyponyms(sub_soup)
                            hyponym.union(additional_hyponym)
                    self._update_cache(sorted(hyponym))
                    return sorted(hyponym)

                except bs4.FeatureNotFound as error:
                    logger.error('An error occurred in the following code segment:')
                    logger.error(''.join(traceback.format_tb(error.__traceback__)))
                except AttributeError as error:
                    logger.error('An AttributeError occurred in the following code segment:')
                    logger.error(''.join(traceback.format_tb(error.__traceback__)))
                except KeyError as error:
                    logger.error('A KeyError occurred in the following code segment:')
                    logger.error(''.join(traceback.format_tb(error.__traceback__)))
                except TypeError as error:
                    logger.error('A TypeError occurred in the following code segment:')
                    logger.error(''.join(traceback.format_tb(error.__traceback__)))
            else:
                hyponym = cleansing.flatten_multidimensional_list([val for val in check_cache.values()])
                return hyponym
